@@ -12,7 +12,7 @@ def main(no_output=False):
     epr_socket = EPRSocket(remote_app_name="Bob")
     with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
         m = epr_socket.create_rsp()[0]
-        print(m.measurement_outcome)
+        print(m.raw_measurement_outcome)
 
     if no_output:
         # Third message since the first two are for init application
