@@ -358,7 +358,7 @@ class Builder:
         )
 
     def _build_cmds_wait_move_epr_to_mem(
-        self, params: EntRequestParams, ent_results_array: Array
+        self, params: EntRequestParams, ent_results_array: Array, role: EPRRole
     ) -> None:
 
         loop_register = self._mem_mgr.get_inactive_register(activate=True)
@@ -373,11 +373,15 @@ class Builder:
                 pair=pair,
             )
 
-            if params.expect_phi_plus:
+            if params.expect_phi_plus and role == EPRRole.CREATE:
                 # Perform Bell corrections
                 bell_state = self._get_raw_bell_state(
                     ent_results_array, loop_reg, bell_state_reg
                 )
+                set_qubit_reg_cmd = ICmd(
+                    instruction=GenericInstr.SET, operands=[qubit_reg, 0]
+                )
+                self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
                 self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
 
             # If it's the last pair, don't move it to a mem qubit
@@ -424,6 +428,7 @@ class Builder:
         params: EntRequestParams,
         ent_results_array: Array,
         tp: EPRType,
+        role: EPRRole,
     ) -> None:
 
         loop_register = self._mem_mgr.get_inactive_register(activate=True)
@@ -439,10 +444,14 @@ class Builder:
             )
             assert tp == EPRType.K or tp == EPRType.R
 
-            if params.expect_phi_plus:
+            if params.expect_phi_plus and role == EPRRole.CREATE:
                 bell_state = self._get_raw_bell_state(
                     ent_results_array, loop_reg, bell_state_reg
                 )
+                set_qubit_reg_cmd = ICmd(
+                    instruction=GenericInstr.SET, operands=[qubit_reg, 0]
+                )
+                self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
                 self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
 
             q_id = qubit_ids.get_future_index(loop_register)
@@ -716,6 +725,7 @@ class Builder:
                         add_new_command=add_new_command,
                         ent_info=ent_info_slice,
                         virtual_address=final_id,
+                        check_initialize=False,
                     )
                     qubits.append(q)
         else:  # generic hardware
@@ -1102,7 +1112,7 @@ class Builder:
     def _build_cmds_move_qubit(self, source: int, target: int) -> None:
         # Moves a qubit from one position to another (assumes that target is free)
         assert target not in [q.qubit_id for q in self._mem_mgr.get_active_qubits()]
-        self._build_cmds_new_qubit(target)
+        self._build_cmds_new_qubit(target, check_initialize=False)
         self._build_cmds_two_qubit(GenericInstr.MOV, source, target)
         self._build_cmds_qfree(source)
 
@@ -1171,7 +1181,24 @@ class Builder:
         commands = [meas_command] + free_commands + outcome_commands  # type: ignore
         self.subrt_add_pending_commands(commands)  # type: ignore
 
-    def _build_cmds_new_qubit(self, qubit_id: int) -> None:
+    def _build_cmds_new_qubit(
+        self, qubit_id: int, check_initialize: bool = True
+    ) -> int:
+        """Returns the ID that is chosen for this qubit."""
+        if (
+            check_initialize
+            and isinstance(self._hardware_config, NVHardwareConfig)
+            and qubit_id != 0
+        ):
+            # NV: only ID 0 can be used for initializing qubits.
+            # If 0 is already used by another qubit, we have to move it to `qubit_id`
+            # and then the new qubit created here will get 0.
+            if (q := self._mem_mgr.get_active_qubit_by_id(0)) is not None:
+                # Move existing qubit away from 0.
+                self._build_cmds_move_qubit(source=0, target=qubit_id)
+                q.qubit_id = qubit_id
+            # Regardless of moving existing qubits or not, the new qubit must have 0.
+            qubit_id = 0
         qubit_reg = self._get_qubit_register()
         self._build_cmds_set_register_value(qubit_reg, qubit_id)
         qalloc_command = ICmd(
@@ -1184,6 +1211,7 @@ class Builder:
         )
         commands = [qalloc_command, init_command]
         self.subrt_add_pending_commands(commands)  # type: ignore
+        return qubit_id
 
     def _build_cmds_init_qubit(self, qubit_id: int) -> None:
         qubit_reg = self._get_qubit_register()
@@ -1307,17 +1335,16 @@ class Builder:
                     # "init old address" + "move to new address", we just have
                     # "init new address"
                     pending_commands = self.subrt_pop_all_pending_commands()
-                    if len(pending_commands) >= 3:
+                    if len(pending_commands) >= 3 and (
                         # Check last 3 commands to see if the qubit was just initialized.
                         # ('type: ignore' since mypy isn't smart enough)
-                        if (
-                            all(isinstance(cmd, ICmd) for cmd in pending_commands[-3:])
-                            and pending_commands[-3].instruction == GenericInstr.SET  # type: ignore
-                            and pending_commands[-2].instruction == GenericInstr.QALLOC  # type: ignore
-                            and pending_commands[-1].instruction == GenericInstr.INIT  # type: ignore
-                        ):
-                            # Update the SET command with the new address.
-                            pending_commands[-3].operands[1] = new_virtual_address  # type: ignore
+                        all(isinstance(cmd, ICmd) for cmd in pending_commands[-3:])
+                        and pending_commands[-3].instruction == GenericInstr.SET  # type: ignore
+                        and pending_commands[-2].instruction == GenericInstr.QALLOC  # type: ignore
+                        and pending_commands[-1].instruction == GenericInstr.INIT  # type: ignore
+                    ):
+                        # Update the SET command with the new address.
+                        pending_commands[-3].operands[1] = new_virtual_address  # type: ignore
                         self.subrt_add_pending_commands(pending_commands)
                     else:
                         self.subrt_add_pending_commands(pending_commands)
@@ -1382,6 +1409,10 @@ class Builder:
             bell_state = self._get_raw_bell_state(
                 ent_results_array, index, bell_state_reg
             )
+            set_qubit_reg_cmd = ICmd(
+                instruction=GenericInstr.SET, operands=[qubit_reg, 0]
+            )
+            self.subrt_add_pending_command(set_qubit_reg_cmd)  # type: ignore
             self._build_cmds_epr_keep_corrections_single_pair(bell_state, qubit_reg)
 
         self._build_cmds_loop_body(
@@ -1494,11 +1525,6 @@ class Builder:
 
         self.subrt_add_pending_commands(wait_cmds)  # type: ignore
 
-        if wait_all and params.expect_phi_plus:
-            self._build_cmds_epr_keep_corrections(
-                qubit_ids_array, ent_results_array, params
-            )
-
     def _build_cmds_epr_create_measure(
         self,
         create_args_array: Array,
@@ -1881,16 +1907,13 @@ class Builder:
 
         if params.post_routine is None and single_comm_qubit:
             self._build_cmds_wait_move_epr_to_mem(
-                params=params, ent_results_array=ent_results_array
+                params=params, ent_results_array=ent_results_array, role=role
             )
 
         # Construct and add NetQASM instructions for post routine
         if params.post_routine:
             self._build_cmds_post_epr(
-                qubit_ids_array,
-                params,
-                ent_results_array,
-                EPRType.K,
+                qubit_ids_array, params, ent_results_array, EPRType.K, role
             )
 
         return qubit_futures, ent_results_array
