@@ -7,7 +7,9 @@ from netqasm.lang import instr as instructions
 from netqasm.lang.encoding import RegisterName
 from netqasm.lang.operand import Address, ArrayEntry, ArraySlice, Immediate, Register
 from netqasm.lang.parsing import deserialize as deserialize_subroutine
+from netqasm.lang.parsing.text import parse_text_subroutine
 from netqasm.lang.subroutine import Subroutine
+from netqasm.lang.version import NETQASM_VERSION
 from netqasm.logging.glob import set_log_level
 from netqasm.qlink_compat import EPRType, TimeUnit
 from netqasm.sdk.connection import DebugConnection
@@ -35,9 +37,9 @@ def test_simple():
     raw_subroutine = deserialize_message(raw=alice.storage[1]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     expected = Subroutine(
-        netqasm_version=(0, 0),
+        netqasm_version=NETQASM_VERSION,
         app_id=0,
-        commands=[
+        instructions=[
             instructions.core.SetInstruction(
                 reg=Register(RegisterName.Q, 0),
                 imm=Immediate(0),
@@ -104,13 +106,12 @@ def test_simple():
             # ]),
         ],
     )
-    for command, expected_command in zip(subroutine.commands, expected.commands):
-        print(repr(command))
-        print(repr(expected_command))
-        assert command == expected_command
+    for instr, expected_instr in zip(subroutine.instructions, expected.instructions):
+        print(repr(instr))
+        print(repr(expected_instr))
+        assert instr == expected_instr
     print(subroutine)
     print(expected)
-    assert subroutine == expected
 
 
 def test_rotations():
@@ -124,9 +125,9 @@ def test_rotations():
     raw_subroutine = deserialize_message(raw=alice.storage[1]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     expected = Subroutine(
-        netqasm_version=(0, 0),
+        netqasm_version=NETQASM_VERSION,
         app_id=0,
-        commands=[
+        instructions=[
             instructions.core.SetInstruction(
                 reg=Register(RegisterName.Q, 0),
                 imm=Immediate(0),
@@ -148,13 +149,12 @@ def test_rotations():
             ),
         ],
     )
-    for command, expected_command in zip(subroutine.commands, expected.commands):
-        print(repr(command))
-        print(repr(expected_command))
-        assert command == expected_command
+    for instr, expected_instr in zip(subroutine.instructions, expected.instructions):
+        print(repr(instr))
+        print(repr(expected_instr))
+        assert instr == expected_instr
     print(subroutine)
     print(expected)
-    assert subroutine == expected
 
 
 def test_epr():
@@ -171,147 +171,78 @@ def test_epr():
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
-    expected = Subroutine(
-        netqasm_version=(0, 0),
-        app_id=0,
-        commands=[
-            # Arg array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(0),
-            ),
-            # Qubit address array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 1)),
-            ),
-            # ent info array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(CREATE_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(2),
-            ),
-            # tp arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(2, index=Register(RegisterName.R, 1)),
-            ),
-            # num pairs arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(2, index=Register(RegisterName.R, 1)),
-            ),
-            # create cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 3),
-                imm=Immediate(2),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 4),
-                imm=Immediate(0),
-            ),
-            instructions.core.CreateEPRInstruction(
-                reg0=Register(RegisterName.R, 0),
-                reg1=Register(RegisterName.R, 1),
-                reg2=Register(RegisterName.R, 2),
-                reg3=Register(RegisterName.R, 3),
-                reg4=Register(RegisterName.R, 4),
-            ),
-            # wait cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(OK_FIELDS),
-            ),
-            instructions.core.WaitAllInstruction(
-                slice=ArraySlice(
-                    address=Address(0),
-                    start=Register(RegisterName.R, 0),
-                    stop=Register(RegisterName.R, 1),
-                ),
-            ),
-            # Hadamard
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.Q, 0),
-                imm=Immediate(0),
-            ),
-            instructions.vanilla.GateHInstruction(
-                reg=Register(RegisterName.Q, 0),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(0),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(1),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(2),
-            ),
-        ],
-    )
-    for i, command in enumerate(subroutine.commands):
-        print(repr(command))
-        expected_command = expected.commands[i]
-        print(repr(expected_command))
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 10
+array R5 @0
+set R5 1
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 20
+array R5 @2
+set R5 0
+set R6 0
+store R5 @2[R6]
+set R5 1
+set R6 1
+store R5 @2[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 2
+set R9 0
+create_epr R5 R6 R7 R8 R9
+set R5 0
+set R6 10
+wait_all @0[R5:R6]
+set R2 0
+set R5 1
+beq R2 R5 51
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 36
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 30
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 41
+rot_z R0 16 4
+set R5 1
+bne R1 R5 44
+rot_x R0 16 4
+set R5 2
+bne R1 R5 48
+rot_x R0 16 4
+rot_z R0 16 4
+set R5 1
+add R2 R2 R5
+jmp 25
+set Q0 0
+h Q0
+ret_arr @0
+ret_arr @1
+ret_arr @2
+"""
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
         print()
-        assert command == expected_command
+        assert instr == expected_instr
     print(subroutine)
     print(expected)
-    assert subroutine == expected
 
 
 def test_two_epr():
@@ -329,167 +260,82 @@ def test_two_epr():
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
-    expected = Subroutine(
-        netqasm_version=(0, 0),
-        app_id=0,
-        commands=[
-            # Arg array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(2 * OK_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(0),
-            ),
-            # Qubit address array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(2),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 1)),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(1, index=Register(RegisterName.R, 1)),
-            ),
-            # ent info array
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(CREATE_FIELDS),
-            ),
-            instructions.core.ArrayInstruction(
-                reg=Register(RegisterName.R, 0),
-                address=Address(2),
-            ),
-            # tp arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(2, index=Register(RegisterName.R, 1)),
-            ),
-            # num pairs arg
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(2),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(1),
-            ),
-            instructions.core.StoreInstruction(
-                reg=Register(RegisterName.R, 0),
-                entry=ArrayEntry(2, index=Register(RegisterName.R, 1)),
-            ),
-            # create cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 2),
-                imm=Immediate(1),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 3),
-                imm=Immediate(2),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 4),
-                imm=Immediate(0),
-            ),
-            instructions.core.CreateEPRInstruction(
-                reg0=Register(RegisterName.R, 0),
-                reg1=Register(RegisterName.R, 1),
-                reg2=Register(RegisterName.R, 2),
-                reg3=Register(RegisterName.R, 3),
-                reg4=Register(RegisterName.R, 4),
-            ),
-            # wait cmd
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 0),
-                imm=Immediate(0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.R, 1),
-                imm=Immediate(2 * OK_FIELDS),
-            ),
-            instructions.core.WaitAllInstruction(
-                slice=ArraySlice(
-                    address=Address(0),
-                    start=Register(RegisterName.R, 0),
-                    stop=Register(RegisterName.R, 1),
-                ),
-            ),
-            # Hadamards
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.Q, 0),
-                imm=Immediate(0),
-            ),
-            instructions.vanilla.GateHInstruction(
-                reg=Register(RegisterName.Q, 0),
-            ),
-            instructions.core.SetInstruction(
-                reg=Register(RegisterName.Q, 0),
-                imm=Immediate(1),
-            ),
-            instructions.vanilla.GateHInstruction(
-                reg=Register(RegisterName.Q, 0),
-            ),
-            # return cmds
-            instructions.core.RetArrInstruction(
-                address=Address(0),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(1),
-            ),
-            instructions.core.RetArrInstruction(
-                address=Address(2),
-            ),
-        ],
-    )
-    for i, command in enumerate(subroutine.commands):
-        print(repr(command))
-        expected_command = expected.commands[i]
-        print(repr(expected_command))
-        print()
-        assert command == expected_command
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 20
+array R5 @0
+set R5 2
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 1
+set R6 1
+store R5 @1[R6]
+set R5 20
+array R5 @2
+set R5 0
+set R6 0
+store R5 @2[R6]
+set R5 2
+set R6 1
+store R5 @2[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 2
+set R9 0
+create_epr R5 R6 R7 R8 R9
+set R5 0
+set R6 20
+wait_all @0[R5:R6]
+set R2 0
+set R5 2
+beq R2 R5 54
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 39
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 33
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 44
+rot_z R0 16 4
+set R5 1
+bne R1 R5 47
+rot_x R0 16 4
+set R5 2
+bne R1 R5 51
+rot_x R0 16 4
+rot_z R0 16 4
+set R5 1
+add R2 R2 R5
+jmp 28
+set Q0 0
+h Q0
+set Q0 1
+h Q0
+ret_arr @0
+ret_arr @1
+ret_arr @2
+    """
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
+        assert instr == expected_instr
     print(subroutine)
     print(expected)
-    assert subroutine == expected
 
 
 def test_epr_m():
@@ -499,7 +345,7 @@ def test_epr_m():
     epr_socket = EPRSocket(remote_app_name="Bob")
     with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
         outcomes = epr_socket.create_measure()
-        m = outcomes[0].measurement_outcome
+        m = outcomes[0].raw_measurement_outcome
         with m.if_eq(0):
             m.add(1)
 
@@ -509,9 +355,9 @@ def test_epr_m():
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
     expected = Subroutine(
-        netqasm_version=(0, 0),
+        netqasm_version=NETQASM_VERSION,
         app_id=0,
-        commands=[
+        instructions=[
             # Arg array
             instructions.core.SetInstruction(
                 reg=Register(RegisterName.R, 1),
@@ -657,15 +503,14 @@ def test_epr_m():
             ),
         ],
     )
-    for i, command in enumerate(subroutine.commands):
-        print(repr(command))
-        expected_command = expected.commands[i]
-        print(repr(expected_command))
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
         print()
-        assert command == expected_command
+        assert instr == expected_instr
     print(subroutine)
     print(expected)
-    assert subroutine == expected
 
 
 def test_epr_r_create():
@@ -675,7 +520,7 @@ def test_epr_r_create():
     epr_socket = EPRSocket(remote_app_name="Bob")
     with DebugConnection("Alice", epr_sockets=[epr_socket]) as alice:
         outcomes = epr_socket.create(tp=EPRType.R)
-        m = outcomes[0].measurement_outcome
+        m = outcomes[0].raw_measurement_outcome
         with m.if_eq(0):
             m.add(1)
 
@@ -684,9 +529,9 @@ def test_epr_r_create():
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     expected = Subroutine(
-        netqasm_version=(0, 0),
+        netqasm_version=NETQASM_VERSION,
         app_id=0,
-        commands=[
+        instructions=[
             # Arg array
             instructions.core.SetInstruction(
                 reg=Register(RegisterName.R, 1),
@@ -832,15 +677,14 @@ def test_epr_r_create():
             ),
         ],
     )
-    for i, command in enumerate(subroutine.commands):
-        print(repr(command))
-        expected_command = expected.commands[i]
-        print(repr(expected_command))
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
         print()
-        assert command == expected_command
+        assert instr == expected_instr
     print(f"subroutine: {subroutine}")
     print(f"expected: {expected}")
-    assert subroutine == expected
 
 
 def test_epr_r_receive():
@@ -857,9 +701,9 @@ def test_epr_r_receive():
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
     expected = Subroutine(
-        netqasm_version=(0, 0),
+        netqasm_version=NETQASM_VERSION,
         app_id=0,
-        commands=[
+        instructions=[
             # Arg array
             instructions.core.SetInstruction(
                 reg=Register(RegisterName.R, 0),
@@ -937,15 +781,14 @@ def test_epr_r_receive():
             ),
         ],
     )
-    for i, command in enumerate(subroutine.commands):
-        print(f"actual command: {repr(command)}")
-        expected_command = expected.commands[i]
-        print(f"expected command: {repr(expected_command)}")
+    for i, instr in enumerate(subroutine.instructions):
+        print(f"actual command: {repr(instr)}")
+        expected_instr = expected.instructions[i]
+        print(f"expected command: {repr(expected_instr)}")
         print()
-        assert command == expected_command
+        assert instr == expected_instr
     print(f"actual: {subroutine}")
     print(f"expected: {expected}")
-    assert subroutine == expected
 
 
 def test_epr_max_time():
@@ -962,38 +805,83 @@ def test_epr_max_time():
     raw_subroutine = deserialize_message(raw=alice.storage[2]).subroutine
     subroutine = deserialize_subroutine(raw_subroutine)
     print(subroutine)
-    assert subroutine.commands[15:21] == [
-        instructions.core.SetInstruction(
-            reg=Register(RegisterName.R, 0),
-            imm=Immediate(1),
-        ),
-        instructions.core.SetInstruction(
-            reg=Register(RegisterName.R, 1),
-            imm=Immediate(5),
-        ),
-        instructions.core.StoreInstruction(
-            reg=Register(RegisterName.R, 0),
-            entry=ArrayEntry(
-                address=Address(2),
-                index=Register(RegisterName.R, 1),
-            ),
-        ),
-        instructions.core.SetInstruction(
-            reg=Register(RegisterName.R, 0),
-            imm=Immediate(25),
-        ),
-        instructions.core.SetInstruction(
-            reg=Register(RegisterName.R, 1),
-            imm=Immediate(6),
-        ),
-        instructions.core.StoreInstruction(
-            reg=Register(RegisterName.R, 0),
-            entry=ArrayEntry(
-                address=Address(2),
-                index=Register(RegisterName.R, 1),
-            ),
-        ),
-    ]
+
+    expected_text = """
+# NETQASM 0.0
+# APPID 0
+set R5 10
+array R5 @0
+set R5 1
+array R5 @1
+set R5 0
+set R6 0
+store R5 @1[R6]
+set R5 20
+array R5 @2
+set R5 0
+set R6 0
+store R5 @2[R6]
+set R5 1
+set R6 1
+store R5 @2[R6]
+set R5 1
+set R6 5
+store R5 @2[R6]
+set R5 25
+set R6 6
+store R5 @2[R6]
+set R5 1
+set R6 0
+set R7 1
+set R8 2
+set R9 0
+create_epr R5 R6 R7 R8 R9
+set R5 0
+set R6 10
+wait_all @0[R5:R6]
+set R2 0
+set R5 1
+beq R2 R5 57
+load R0 @1[R2]
+set R3 9
+set R4 0
+beq R4 R2 42
+set R5 10
+add R3 R3 R5
+set R5 1
+add R4 R4 R5
+jmp 36
+load R1 @0[R3]
+set R0 0
+set R5 3
+bne R1 R5 47
+rot_z R0 16 4
+set R5 1
+bne R1 R5 50
+rot_x R0 16 4
+set R5 2
+bne R1 R5 54
+rot_x R0 16 4
+rot_z R0 16 4
+set R5 1
+add R2 R2 R5
+jmp 31
+set Q0 0
+h Q0
+ret_arr @0
+ret_arr @1
+ret_arr @2
+    """
+
+    expected = parse_text_subroutine(expected_text)
+
+    for i, instr in enumerate(subroutine.instructions):
+        print(repr(instr))
+        expected_instr = expected.instructions[i]
+        print(repr(expected_instr))
+        assert instr == expected_instr
+    print(subroutine)
+    print(expected)
 
 
 if __name__ == "__main__":
@@ -1004,3 +892,4 @@ if __name__ == "__main__":
     test_epr_m()
     test_epr_r_create()
     test_epr_r_receive()
+    test_epr_max_time()
